#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG3 \
//Spr J[\

//---------------------------------------------------------------------------
//
#define BG_CX					8
#define BG_CY					8
#define BG_SCREEN_CX			30
#define BG_SCREEN_CY			20
#define BG_WORLD_SCREEN_CX		32
#define BG_WORLD_SCREEN_CY		32
#define BG_WORLD_SCREEN_WCX		256
#define BG_WORLD_SCREEN_WCY		256

//is
#define STATE_INIT				1
#define STATE_START				2
#define STATE_TURN				3
#define STATE_END_JUDGE			5
#define STATE_END				6

//
#define HAND_TYPE_PLAYER		1
#define HAND_TYPE_COMPUTER		2
#define HAND_STAT_NORMAL		1
#define HAND_STAT_WIN			2
#define HAND_STAT_LOSE			3
#define HAND_MAX_CNT			4
#define HAND_MAX_CARD_CNT		20
#define HAND_MAX_PASS_CNT		2

//
#define CARD_NUM				13
#define CARD_SUITS				4
#define CARD_TOTAL_CNT			(CARD_NUM * CARD_SUITS)
#define CARD_SHUFFLE_CNT		(CARD_TOTAL_CNT * 5)

//̓L[̃EFCg
#define KEY_REPEAT_TIME			500
#define TIMER_INTMS				10

//---------------------------------------------------------------------------
typedef struct {
	u8 Chr;
	u8 Map;
	u8 Col;
} ST_BG;

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	s16 minX;							//ړ
	s16 maxX;
	s16 minY;
	s16 maxY;
	s16 moveX;
	s16 moveY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
	u8  keyNow;							//ړp
	u8  keyOld;							//1O̓
} ST_CURSOR;

typedef struct {
	u32 clockTick;
	u32 wait;
	u32 repeat;
} ST_TIMER;

typedef struct {
	u8  num;
	u8  type;
	u8  stat;
	s8  passCnt;
	s8  cardCnt;						//D̐
	u8  card[HAND_MAX_CARD_CNT];
} ST_HAND;

//---------------------------------------------------------------------------
ST_BG     BG3;
ST_CURSOR Cursor;
ST_TIMER  Timer;

u8        Deck [CARD_TOTAL_CNT];		//J[h̃fbL
ST_HAND   Hand [HAND_MAX_CNT];			//vC + COM1`3
bool      Field[CARD_SUITS][CARD_NUM];	//tB[h̃J[hL

u8  State;								//istO
s8  Target;								//
s8  ClearCnt;
//---------------------------------------------------------------------------
void StateInit();
void StateStart();
void StateTurn();
void StateEndJudge();
void StateEnd();

void InitWS();
void InitBG();
void InitSprite();
void InitCursor();
void InitTimer();
void InitDeck();
void InitHand();
void InitField();
void IRQUserHandler();

void DrawFrame();
void DrawCursor();
void DrawField(u8 num);
void DrawCard(u8 x, u8 y, u8 num, bool isMark);
void DrawCardList(ST_HAND* p, u8 x, u8 y);
void DrawNumber(u8 x, u8 y, u8 num);
void DrawHandStatusAll();
void DrawHandStatus(ST_HAND* p);

void CardShuffle();
void CardSwap(u8 src, u8 dst);
void CardDeal();
u8   CardGetMark(u8 num);
u8   CardGetNumber(u8 num);

void HandSortAll();
void HandSort(ST_HAND* p);
void HandSwap(ST_HAND* p, u8 src, u8 dst);
void HandPickup7();
void HandDeal(ST_HAND* p, u8 num);
bool HandChkDeal(u8 num);

void FieldSet(u8 num);

void ProcPlayer();
void ProcComputer();
void ProcKeyInput();

bool ChkMoveCursor();
void MoveCursor();
bool WaitTimer();
bool RepeatKey();

u8   GetCursorxToNum();
//---------------------------------------------------------------------------
int main(void)
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:      StateInit();     break;
			case STATE_START:     StateStart();    break;
			case STATE_TURN:      StateTurn();     break;
			case STATE_END_JUDGE: StateEndJudge(); break;
			case STATE_END:       StateEnd();      break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	KeyInit();

	InitWS();
	InitBG();
	InitSprite();
	InitCursor();
	InitTimer();
	InitDeck();
	InitHand();
	InitField();

	Target   = 0;
	ClearCnt = 0;

	SetMode(MODE_0 | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_START;
}
//---------------------------------------------------------------------------
//EFCgRg[
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG3 \

	BG3.Chr = 1;
	BG3.Map = 0;
	BG3.Col = 0;

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 0);

	BGSetPal16  (BG3.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG3.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();
	SpriteSetPal16((u16*)&cursor_pal_bin);
	SpriteSetData ((u16*)&cursor_img_bin, CURSOR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitCursor()
{
	ST_CURSOR* cp = &Cursor;

	cp->x       =  1;
	cp->y       = 19;
	cp->cx      =  8;
	cp->cy      =  8;
	cp->minX    =  1;			//ړ
	cp->maxX    = 25;
	cp->minY    = 19;
	cp->maxY    = 19;
	cp->moveX   =  2;
	cp->moveY   =  2;
	cp->sprNo   =  0;
	cp->sprTile =  0;
	cp->sprSize = SP_SIZE_16;
	cp->sprForm = SP_SQUARE;

	SpriteSetSize(Cursor.sprNo, Cursor.sprSize, Cursor.sprForm, SP_COLOR_16);
	SpriteSetChr (Cursor.sprNo, Cursor.sprTile);
}
//---------------------------------------------------------------------------
void InitTimer()
{
	Timer.wait      = 0;
	Timer.repeat    = 0;
	Timer.clockTick = 0;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;

	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void InitDeck()
{
	s16 i;
	for(i=0; i<CARD_TOTAL_CNT; i++)
	{
		Deck[i] = i;
	}
}
//---------------------------------------------------------------------------
void InitHand()
{
	s16 i;
	for(i=0; i<HAND_MAX_CNT; i++)
	{
		Hand[i].num     = i;
		Hand[i].type    = HAND_TYPE_COMPUTER;
		Hand[i].stat    = HAND_STAT_NORMAL;
		Hand[i].passCnt = HAND_MAX_PASS_CNT;
		Hand[i].cardCnt = 0;
	}

	Hand[0].type = HAND_TYPE_PLAYER;
}
//---------------------------------------------------------------------------
void InitField()
{
	s16 i,j;
	for(i=0; i<CARD_SUITS; i++)
	{
		for(j=0; j<CARD_NUM; j++)
		{
			Field[i][j] = FALSE;
		}
	}
}
//---------------------------------------------------------------------------
void StateStart()
{
	DrawFrame();
	DrawCursor();

	CardShuffle();
	CardDeal();

	HandPickup7();
	HandSort(&Hand[0]);		//vĈݎD̃\[g܂
	DrawHandStatusAll();

	State = STATE_TURN;
}
//---------------------------------------------------------------------------
void CardShuffle()
{
	u8  src, dst;
	s16 i;

	for(i=0; i<CARD_SHUFFLE_CNT; i++)
	{
		src = rand() % CARD_TOTAL_CNT;
		dst = rand() % CARD_TOTAL_CNT;

		CardSwap(src, dst);
	}
}
//---------------------------------------------------------------------------
void CardSwap(u8 src, u8 dst)
{
	u8 s = Deck[src];
	u8 d = Deck[dst];

	Deck[src] = d;
	Deck[dst] = s;
}
//---------------------------------------------------------------------------
void CardDeal()
{
	s16 i;
	s8  t ,c;

	for(i=0; i<CARD_TOTAL_CNT; i++)
	{
		t = i % HAND_MAX_CNT;		//z鑊肵܂
		c = Hand[t].cardCnt;

		Hand[t].card[c] = Deck[i];
		Hand[t].cardCnt++;
	}
}
//---------------------------------------------------------------------------
void HandPickup7()
{
	ST_HAND* p;
	s16 hc;		//HandCnt
	s16 i;

	for(hc=0; hc<HAND_MAX_CNT; hc++)
	{
		p = &Hand[hc];

		for(i=0; i<p->cardCnt; i++)
		{
			if( CardGetNumber(p->card[i]) == '7' )
			{
				//J[ho܂
				HandDeal(p, i);

				//7̃Xy[h`FbN܂
				if( CardGetMark(p->card[i]) == 'S' )
				{
					//oԂ̐s肵܂
					Target = p->num;
				}

				i--; //oJ[ḧʒuɁAʃJ[h
			}
		}
	}
}
//---------------------------------------------------------------------------
//tB[hɃJ[hu܂
void HandDeal(ST_HAND* p, u8 num)
{
	DrawField(p->card[num]);
	FieldSet (p->card[num]);

	p->cardCnt--;
	p->card[num] = p->card[p->cardCnt];

	//D̍XV܂
	if(p->type == HAND_TYPE_PLAYER)
	{
		HandSort(p);
		Cursor.maxX -= Cursor.moveX;
	}
	DrawHandStatus(p);
}
//---------------------------------------------------------------------------
void FieldSet(u8 srcNum)
{
	u8   mark = srcNum / CARD_NUM;
	u8   num  = srcNum % CARD_NUM;

	Field[mark][num] = TRUE;
}
//---------------------------------------------------------------------------
void DrawField(u8 num)
{
	u8 x = 3 + (num % CARD_NUM) * 2;
	u8 y = 7 + (num / CARD_NUM) * 2;

	DrawCard(x, y, num, FALSE);
}
//---------------------------------------------------------------------------
void HandSortAll()
{
	s16 hc;		//HandCnt

	for(hc=0; hc<HAND_MAX_CNT; hc++)
	{
		HandSort(&Hand[hc]);
	}
}
//---------------------------------------------------------------------------
void HandSort(ST_HAND* p)
{
	s16 i,j;

	//ou\[g
	for(i=0; i<p->cardCnt; i++)
	{
		for (j=0; j<p->cardCnt-1; j++)
		{
			if(p->card[j+1] < p->card[j])
			{
				HandSwap(p, j+1, j);
			}
		}
	}
}
//---------------------------------------------------------------------------
void HandSwap(ST_HAND* p, u8 src, u8 dst)
{
	u8 s = p->card[src];
	u8 d = p->card[dst];

	p->card[src] = d;
	p->card[dst] = s;
}
//---------------------------------------------------------------------------
void DrawFrame()
{
	BGDrawCls(BG3.Map, ' ');

	BGDrawStr(BG3.Map,  1,  7, (u8*)"S");	//Xy[h
	BGDrawStr(BG3.Map,  1,  9, (u8*)"H");	//n[g
	BGDrawStr(BG3.Map,  1, 11, (u8*)"D");	//_C
	BGDrawStr(BG3.Map,  1, 13, (u8*)"C");	//N[o

	BGDrawStr(BG3.Map,  1, 15, (u8*)"Player");
	BGDrawStr(BG3.Map, 11, 15, (u8*)"Pass:");
	BGDrawStr(BG3.Map, 21, 15, (u8*)"Card:");

	BGDrawStr(BG3.Map,  1,  2, (u8*)"Com1 ");
	BGDrawStr(BG3.Map,  1,  4, (u8*)"Pass:");
	BGDrawStr(BG3.Map,  1,  5, (u8*)"Card:");

	BGDrawStr(BG3.Map, 11,  2, (u8*)"Com2 ");
	BGDrawStr(BG3.Map, 11,  4, (u8*)"Pass:");
	BGDrawStr(BG3.Map, 11,  5, (u8*)"Card:");

	BGDrawStr(BG3.Map, 21,  2, (u8*)"Com3 ");
	BGDrawStr(BG3.Map, 21,  4, (u8*)"Pass:");
	BGDrawStr(BG3.Map, 21,  5, (u8*)"Card:");
}
//---------------------------------------------------------------------------
void DrawCursor()
{
	s16 x  = Cursor.x * Cursor.cx;
	s16 y  = Cursor.y * Cursor.cy;
	u16 no = Cursor.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
void DrawHandStatusAll()
{
	s16 hc;		//HandCnt

	for(hc=0; hc<HAND_MAX_CNT; hc++)
	{
		DrawHandStatus(&Hand[hc]);
	}
}
//---------------------------------------------------------------------------
void DrawHandStatus(ST_HAND* p)
{
	u8 x, y;
	
	switch(p->type)
	{
		case HAND_TYPE_PLAYER:
			DrawCardList(p, 1, 17);
			DrawNumber(16, 15, p->passCnt);
			DrawNumber(27, 15, p->cardCnt);
			break;

		case HAND_TYPE_COMPUTER:
			x = 7 + 10 * (p->num - 1);
			y = 4;

			DrawNumber(x, y+0, p->passCnt);
			DrawNumber(x, y+1, p->cardCnt);
			break;

		default:
			DebugMessage((u8*)"Err: DrawHandStatus type");
			break;
	}
}
//---------------------------------------------------------------------------
void DrawCardList(ST_HAND* p, u8 x, u8 y)
{
	BGDrawBox(BG3.Map, x, y, 28, 2, ' ');

	s16 i;
	for(i=0; i<p->cardCnt; i++)
	{
		DrawCard(x, y, p->card[i], TRUE);
		x += 2;
	}
}
//---------------------------------------------------------------------------
void DrawNumber(u8 x, u8 y, u8 num)
{
	char msg[20];
	bprintf(msg, "%2d", num);
	BGDrawStr(BG3.Map, x, y, (u8*)msg);
}
//---------------------------------------------------------------------------
void DrawCard(u8 x, u8 y, u8 num, bool isMark)
{
	u8 dispMark = CardGetMark(num);
	u8 dispNum  = CardGetNumber(num);

	//\`߂܂
	if(isMark == TRUE)
	{
		BGDrawChar(BG3.Map, x, y  , BG3.Col, dispMark);
		BGDrawChar(BG3.Map, x, y+1, BG3.Col, dispNum);
	}
	else
	{
		BGDrawChar(BG3.Map, x, y, BG3.Col, dispNum);
	}
}
//---------------------------------------------------------------------------
u8 CardGetMark(u8 num)
{
	u8 ret;
	switch(num / CARD_NUM)
	{
		case  0: ret = 'S'; break;
		case  1: ret = 'H'; break;
		case  2: ret = 'D'; break;
		case  3: ret = 'C'; break;
		default: ret = ' '; break;		//G[iȂ
	}

	return ret;
}
//---------------------------------------------------------------------------
u8 CardGetNumber(u8 num)
{
	u8 ret;
	switch(num % CARD_NUM)
	{
		case  0: ret = 'A'; break;
		case  9: ret = 'T'; break;		//Ten(10)
		case 10: ret = 'J'; break;
		case 11: ret = 'Q'; break;
		case 12: ret = 'K'; break;
		default: ret = (num % CARD_NUM) + 0x30 + 1; break;
	}

	return ret;
}
//---------------------------------------------------------------------------
void StateTurn()
{
	ST_HAND* p = &Hand[Target];

	if(p->stat != HAND_STAT_NORMAL)
	{
		State = STATE_END_JUDGE;
		return;
	}


	if(p->type == HAND_TYPE_PLAYER)
	{
		ProcPlayer();
	}
	else if(p->type == HAND_TYPE_COMPUTER)
	{
		ProcComputer();
	}
}
//---------------------------------------------------------------------------
void ProcPlayer()
{
	ProcKeyInput();
}
//---------------------------------------------------------------------------
void ProcComputer()
{
	ST_HAND* p = &Hand[Target];

	s16 i;
	for(i=0; i<p->cardCnt; i++)
	{
		if( HandChkDeal(p->card[i]) == TRUE )
		{
			HandDeal(p, i);
			goto EndProc;
		}
	}

	p->passCnt--;

EndProc:

	DrawHandStatus(p);
	State = STATE_END_JUDGE;
}
//---------------------------------------------------------------------------
bool HandChkDeal(u8 srcNum)
{
	u8   mark = srcNum / CARD_NUM;
	u8   num  = srcNum % CARD_NUM;
	s8   x;

	//EɒT܂
	for(x=6+1; ; x++)
	{
		if((x % CARD_NUM)==0)
		{
			x -= CARD_NUM;
		}
		if(x == num)
		{
			return TRUE;
		}
		if(Field[mark][x] == FALSE)
		{
			break;
		}
	}

	//ɒT܂
	for(x=6-1; ; x--)
	{
		if((x % CARD_NUM)==0)
		{
			x += CARD_NUM;
		}
		if(x == num)
		{
			return TRUE;
		}
		if(Field[mark][x] == FALSE)
		{
			break;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
u8 GetCursorxToNum()
{
	ST_CURSOR* p = &Cursor;
	return p->x == 1 ? 0 : (p->x / 2);
}
//---------------------------------------------------------------------------
void StateEndJudge()
{
	ST_HAND* p = &Hand[Target];

	if(p->stat != HAND_STAT_NORMAL)
	{
		goto EndState;
	}

	if(p->passCnt == 0)
	{
		p->stat = HAND_STAT_LOSE;

		while(p->cardCnt > 0)
		{
			HandDeal(p, 0);
		}
		DrawHandStatus(p);

		ClearCnt++;
		if(ClearCnt >= HAND_MAX_CNT)
		{
			State = STATE_END;
			return;
		}
	}
	else if(p->cardCnt == 0)
	{
		p->stat = HAND_STAT_WIN;

		ClearCnt++;
		if(ClearCnt >= HAND_MAX_CNT)
		{
			State = STATE_END;
			return;
		}
	}

EndState:

	Target = (Target + 1) % HAND_MAX_CNT;
	State  = STATE_TURN;
}
//---------------------------------------------------------------------------
void StateEnd()
{
	
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	//^C}[EFCg
	if( WaitTimer() == TRUE )
	{
		return;
	}

	ST_CURSOR* p = &Cursor;
	p->keyNow    = KeyInput();

	//s[h
	if( RepeatKey() == TRUE )
	{
		return;
	}

	if( ChkMoveCursor() == TRUE )
	{
		MoveCursor();
		DrawCursor();
	}

	ST_HAND* h = &Hand[Target];

	if(p->keyNow & KEY_A)		//J[hI
	{
		u8 num     = GetCursorxToNum();
		u8 cardNum = h->card[num];

		if( HandChkDeal(cardNum) == TRUE )
		{
			HandDeal(h, num);
			HandSort(h);
			DrawHandStatus(h);

			State = STATE_END_JUDGE;
		}
	}
	else if(p->keyNow & KEY_B)	//pX
	{
		h->passCnt--;
		DrawHandStatus(h);

		State = STATE_END_JUDGE;
	}
}
//---------------------------------------------------------------------------
//L[s[hi`^Oh~j
bool RepeatKey()
{
	ST_CURSOR* p = &Cursor;
	ST_TIMER*  t = &Timer;

	if(p->keyNow && p->keyNow == p->keyOld)
	{
		if(t->clockTick - t->repeat < KEY_REPEAT_TIME)
		{
			return TRUE;
		}

		return FALSE;
	}

	p->keyOld = p->keyNow;
	t->repeat = t->clockTick;

	return FALSE;
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
bool ChkMoveCursor()
{
	ST_CURSOR* p = &Cursor;

	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->keyNow & KEY_UP   ) ty1 -= p->moveY;
	if(p->keyNow & KEY_DOWN ) ty1 += p->moveY;
	if(p->keyNow & KEY_LEFT ) tx1 -= p->moveX;
	if(p->keyNow & KEY_RIGHT) tx1 += p->moveX;

	if(tx1 < p->minX || ty1 < p->minY) return FALSE;
	if(tx1 > p->maxX || ty1 > p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
void MoveCursor()
{
	ST_CURSOR* p = &Cursor;

	if(p->keyNow & KEY_UP   ) p->y -= p->moveY;
	if(p->keyNow & KEY_DOWN ) p->y += p->moveY;
	if(p->keyNow & KEY_LEFT ) p->x -= p->moveX;
	if(p->keyNow & KEY_RIGHT) p->x += p->moveX;
}
//---------------------------------------------------------------------------
bool WaitTimer()
{
	if(Timer.clockTick - Timer.wait < TIMER_INTMS)
	{
		return TRUE;
	}

	Timer.wait = Timer.clockTick;
	return FALSE;
}
